@@ -1,15 +1,21 @@
 package Munin::Master::LimitsOld;
 # -*- perl  -*-
 
-=begin comment
+=head1 NAME
+
+Munin::Master::LimitsOld - Abstract base class for workers.
+
+=head1 SYNOPSIS
 
 This is Munin::Master::LimitsOld, a minimal package shell to make
 munin-limits modular (so it can be loaded persistently in a daemon for
 example) without making it object oriented yet.  The non-'old' module
-will feature propper object orientation like munin-update and will
+will feature proper object orientation like munin-update and will
 have to wait until later.
 
 
+=begin comment
+
 Copyright (C) 2004-2009 Jimmy Olsen
 
 This program is free software; you can redistribute it and/or
@@ -26,8 +32,6 @@ You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 
-$Id$
-
 =end comment
 
 =cut
@@ -103,7 +107,7 @@ sub limits_startup {
     @always_send = qw{ok warning critical unknown} if $force;
 
     munin_readconfig_base($conffile);
-    # XXX: check if it does actualy need that part
+    # XXX: check if it does actually need that part
     $config = munin_readconfig_part('datafile', 0);
 
     logger_open($config->{'logdir'});
@@ -369,29 +373,26 @@ sub process_service {
 		my ($previous_updated_timestamp, $previous_updated_value) = @{ $state->{value}{"$rrd_filename:42"}{previous} || [ ] };
 
 		my $heartbeat = 600; # XXX - $heartbeat is a fixed 10 min (2 runs of 5 min).
-		if (! $field->{type} || $field->{type} eq "GAUGE") {
-			$value = $current_updated_value;
-		} elsif (! defined $current_updated_value || ! defined $previous_updated_value || $current_updated_timestamp == $previous_updated_timestamp) {
-			# No derive computing possible. Report unknown.
+		if (! defined $current_updated_value || $current_updated_value eq "U") {
+			# No value yet. Report unknown.
 			$value = "U";
 		} elsif (time > $current_updated_timestamp + $heartbeat) {
-			# Current value is too old. Report unknown. 
+			# Current value is too old. Report unknown.
 			$value = "U";
-		} elsif ($current_updated_timestamp > $previous_updated_timestamp + $heartbeat) {
-			# Old value is too old. Report unknown. 
+		} elsif (! $field->{type} || $field->{type} eq "GAUGE") {
+			# Non-compute up-to-date value.
+			$value = $current_updated_value;
+		} elsif (! defined $previous_updated_value || $previous_updated_value eq "U") {
+			# No derive computing possible. Report unknown.
 			$value = "U";
-		} elsif ($field->{type} eq "COUNTER" && $current_updated_value < $previous_updated_value) {
-			# COUNTER never decrease. Report unknown.
-			$value = "U";
-		} elsif ($current_updated_value eq "U") {
-			# The current value is unknown. Report unknown.
+		} elsif ($current_updated_timestamp == $previous_updated_timestamp || $current_updated_timestamp > $previous_updated_timestamp + $heartbeat) {
+			# Old value does not exist or is too old. Report unknown.
 			$value = "U";
 		} elsif ($field->{type} eq "ABSOLUTE") {
-			# The previous value is unimportant, as if ABSOLUTE, the counter is reset anytime the value is read
+			# The previous value is unimportant, as if ABSOLUTE, the counter is reset every time the value is read
 			$value = $current_updated_value / ($current_updated_timestamp - $previous_updated_timestamp);
-		} elsif ($previous_updated_value eq "U") {
-			# The previous value is unknown.
-			# Report unknown, as we are not ABSOLUTE
+		} elsif ($field->{type} eq "COUNTER" && $current_updated_value < $previous_updated_value) {
+			# COUNTER never decrease. Report unknown.
 			$value = "U";
 		} else {
 			# Everything is ok for DERIVE/COUNTER
@@ -401,11 +402,15 @@ sub process_service {
 	}
 
         # De-taint.
-        if (!defined $value || $value eq "U") {
+        if ($value eq "U") {
             $value = "unknown";
         }
         else {
-            $value = sprintf "%.2f", $value;
+            my $formatted_value = sprintf "%.2f", $value;
+            if (($formatted_value == 0) && !looks_like_number($value)) {
+                warn "Failed to interpret expected numeric value of field '$fname' (host '$host'): '$value'";
+            }
+            $value = $formatted_value;
         }
 
         # Some fields that are nice to have in the plugin output
@@ -650,6 +655,16 @@ sub generate_service_message {
     $hash->{'ufields'}  = join " ", @{$stats{'unknown'}};
     $hash->{'fofields'} = join " ", @{$stats{'foks'}};
     $hash->{'ofields'}  = join " ", @{$stats{'ok'}};
+    # The "fofields" (datasets that changed state from "failed" to "OK") can be empty under certain
+    # legitimate circumstances (e.g. "munin-limits --force" sends messages also for unchanged "OK"
+    # states).  But we may never allow the fourth output field for NSCA to be empty - otherwise the
+    # recipient (nagios/icinga) cannot determine, which fields are affected (and thus which test
+    # succeeded).  Thus we need to make sure, that "fofields" is always defined (since our
+    # self-made trivial template language does not support expressions like "fofields || ofields").
+    # Here "ofields" is a reasonable fallback value: it contains all datasets with status "OK".
+    if ($hash->{'fofields'} eq '') {
+        $hash->{'fofields'} = $hash->{'ofields'};
+    }
     $hash->{'numcfields'}  = scalar @{$stats{'critical'}};
     $hash->{'numwfields'}  = scalar @{$stats{'warning'}};
     $hash->{'numufields'}  = scalar @{$stats{'unknown'}};
@@ -695,7 +710,7 @@ sub generate_service_message {
         if (!$hash->{'state_changed'} and !$obsess) {
             next;    # No need to send notification
         }
-        DEBUG "[DEBUG] state has changed, notifying $c";
+        INFO("[INFO] state has changed, notifying $c");
         my $precmd = munin_get($contactobj, "command", undef);
         if(!defined $precmd) {
             WARN("[WARNING] Missing command option for contact $c; skipping");
@@ -752,6 +767,12 @@ sub generate_service_message {
             } else { # child
                 close $w;
                 open(STDIN, "<&", $r);
+                # We want to close stdout before calling the send command. This prevents the
+                # notification program (e.g. "send_nsca") to write irrelevant status messages
+                # to stdout and thus trigger notification emails (e.g. via cron).
+                # See https://github.com/munin-monitoring/munin/issues/382
+                # and https://bugs.debian.org/cgi-bin/bugreport.cgi?bug=291168.
+                close(STDOUT);
                 exec($cmd) or WARN "[WARNING] Failed to exec for contact $c in pid $$";
                 exit;
             }
